#define _FILE_OFFSET_BITS 64
#define _GNU_SOURCE

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <sys/time.h>
#include <time.h>
#include "kerncompat.h"
#include "radix-tree.h"
#define ACP_VERSION "0.6"

#define RADIX_TAG_PENDING 0
#define RADIX_TAG_DIR_PENDING 1
#define RADIX_TAG_READ_PENDING 1

static int read_one_fd(int fd, char *fname, struct stat *st);

static char *buf;
static int bufsize;
static int total_reada = 0;
static int total_stata = 0;
static int total_open = 0;
static int max_stata = 512;
static int max_reada = 8 * 1024 * 1024;
static loff_t total_bytes_read = 0;

static int read_whole_dir_tree = 0;
static int sort_by_bmap = 0;
static int inline_files = 0;
static int verbose = 0;
static int metadata_only = 0;

struct node {
	unsigned long ino;
	loff_t isize;
	int d_type;
	char fname[];
};

struct reada_node {
	unsigned long ino;
	unsigned long seq;
	struct stat st;
	int fd;
	loff_t reada;
	char fname[];
};

int add_node(struct radix_tree_root *radix, unsigned long ino, char *dir,
	     char *fname, int d_type)
{
	int dirlen = strlen(dir);
	int namelen = strlen(fname);
	/* one for the null, one for the slash */
	struct node *rec = malloc(sizeof(*rec) + dirlen + namelen + 2);
	int ret;
	if (!rec)
		return -ENOMEM;
	strncpy(rec->fname, dir, dirlen);
	rec->fname[dirlen] = '/';
	strncpy(rec->fname + dirlen + 1, fname, namelen);
	rec->fname[dirlen + namelen + 1] = '\0';
	rec->ino = ino;
	rec->d_type = d_type;
	ret = radix_tree_insert(radix, ino, rec);
	if (ret == -EEXIST) {
		ret = 0;
		goto fail;
	}
	if (ret)
		goto fail;

	if (d_type == DT_DIR)
		radix_tree_tag_set(radix, ino, RADIX_TAG_DIR_PENDING);
	else
		radix_tree_tag_set(radix, ino, RADIX_TAG_PENDING);

	return 0;
fail:
	free(rec);
	return ret;
}

int add_reada_node(struct radix_tree_root *radix, int fd, char *fname,
		   struct stat *st, unsigned long block0)
{
	int namelen = strlen(fname);
	struct reada_node *rec = malloc(sizeof(*rec) + namelen + 1);
	int ret;

	if (!rec)
		return -ENOMEM;
	strcpy(rec->fname, fname);
	rec->fd = fd;
	memcpy(&rec->st, st, sizeof(*st));
	rec->ino = st->st_ino;
	rec->reada = 0;
	if (sort_by_bmap)
		rec->seq = block0;
	else
		rec->seq = rec->ino;
	ret = radix_tree_insert(radix, rec->seq, rec);
	if (ret == -EEXIST) {
		ret = 0;
		goto fail;
	}
	if (ret)
		goto fail;
	radix_tree_tag_set(radix, rec->seq, RADIX_TAG_PENDING);
	total_stata++;
	return 0;
fail:
	free(rec);
	return ret;
}

static int read_tree(struct radix_tree_root *radix, char *root,
		     unsigned long limit)
{
	DIR *dir;
	struct dirent *dirent;
	int ret;
	int namelen;

	dir = opendir(root);

	if (!dir) {
		perror("opendir");
		exit(1);
	}
	while((dirent = readdir(dir))) {
		namelen = strlen(dirent->d_name);
		if (namelen <= 2 &&
		    !strcmp(dirent->d_name, ".."))
			continue;
		if (namelen <= 1 &&
		    !strcmp(dirent->d_name, "."))
			continue;
		ret = add_node(radix, dirent->d_ino, root, dirent->d_name,
			       dirent->d_type);
		if (ret) {
			fprintf(stderr,
				"unable to create record for %s/%s err %d\n",
				root, dirent->d_name, ret);
		}
	}
	closedir(dir);
	return 0;
}

static void stat_to_dtype(struct node *n, struct stat *st)
{
	if (S_ISDIR(st->st_mode))
		n->d_type = DT_DIR;
	else if (S_ISREG(st->st_mode))
		n->d_type = DT_REG;
	else if (S_ISLNK(st->st_mode))
		n->d_type = DT_LNK;
	else if (S_ISCHR(st->st_mode))
		n->d_type = DT_CHR;
	else if (S_ISBLK(st->st_mode))
		n->d_type = DT_BLK;
	else if (S_ISFIFO(st->st_mode))
		n->d_type = DT_FIFO;
	else if (S_ISSOCK(st->st_mode))
		n->d_type = DT_SOCK;
}

static int stat_one(struct node *n)
{
	struct stat st;
	int ret;
	ret = stat(n->fname, &st);
	if (ret != 0)
		return ret;
	stat_to_dtype(n, &st);
	return 0;
}

static int read_next_dir(struct radix_tree_root *radix)
{
	struct node *n;
	int ret;
	ret = radix_tree_gang_lookup_tag(radix,
					 (void **)&n,
					 0, 1,
					 RADIX_TAG_DIR_PENDING);
	if (!ret)
		return 0;
	if (n->d_type == DT_UNKNOWN) {
		stat_one(n);
	}
	if (n->d_type == DT_DIR) {
		read_tree(radix, n->fname, 0);
	} else {
		radix_tree_tag_set(radix, n->ino,
				   RADIX_TAG_PENDING);
	}
	radix_tree_tag_clear(radix, n->ino,
			     RADIX_TAG_DIR_PENDING);
	return 1;
}

static int read_one_fd(int fd, char *fname, struct stat *st)
{
	ssize_t cur = 0;
	loff_t total_read = 0;

	if (metadata_only) {
		close(fd);
		return 0;
	}
	if (verbose) {
		fprintf(stderr, "reading %s inum %lu\n", fname, st->st_ino);
	}
	while(1) {
		cur = read(fd, buf, bufsize);
		if (cur < 0) {
			perror("read");
			exit(1);
		}
		if (cur == 0) {
			if (total_read < st->st_size) {
				fprintf(stderr, "error, short read on %s\n",
					fname);
			}
			break;
		}
		total_read += cur;
	}
	total_open--;
	close(fd);
	total_bytes_read += total_read;
	return 0;
}

static int run_reada(struct radix_tree_root *radix,
		     struct radix_tree_root *reads_radix)
{
	int ret;
	struct reada_node *gang[16];
	struct reada_node *n;
	struct reada_node *p;
	int i;
	int num_run = 0;
	unsigned long cur_reada;
	static unsigned long reads_index = 0;

	while(1) {
		if (total_reada >= max_reada)
			break;
		ret = radix_tree_gang_lookup_tag(radix,
						 (void **)gang,
						 0, ARRAY_SIZE(gang),
						 RADIX_TAG_PENDING);
		if (!ret)
			break;
		for (i = 0; i < ret && total_reada < max_reada; i++) {
			unsigned long ins;
			n = gang[i];
			if (read_whole_dir_tree)
				ins = n->seq;
			else
				ins = reads_index;
			radix_tree_insert(reads_radix, ins, n);
			p = radix_tree_delete(radix, n->seq);
			BUG_ON(p != n);
			n->seq = ins;
			reads_index++;
			total_stata -= 1;
			cur_reada = n->st.st_size;
			if (cur_reada > max_reada - total_reada)
				cur_reada = max_reada - total_reada;
			total_reada += cur_reada;
			n->reada = cur_reada;
			if (verbose) {
				fprintf(stderr, "readahead %s %lu\n", n->fname,
					n->ino);
			}
			readahead(n->fd, 0, cur_reada);
			num_run++;
		}
	}
	return num_run;
}

static int run_reads(struct radix_tree_root *radix)
{
	int ret;
	struct reada_node *gang[16];
	struct reada_node *n;
	struct reada_node *p;
	int i;
	int num_run = 0;
	while(1) {
		ret = radix_tree_gang_lookup(radix, (void **)gang,
					     0, ARRAY_SIZE(gang));
		if (!ret)
			break;
		for (i = 0; i < ret; i++) {
			n = gang[i];
			read_one_fd(n->fd, n->fname, &n->st);
			total_reada -= n->reada;
			num_run++;
			p = radix_tree_delete(radix, n->seq);
			BUG_ON(p != n);
			free(n);
		}
	}
	return num_run;
}

static int reada_one_fd(struct radix_tree_root *radix,
			struct radix_tree_root *reads_radix,
			int fd, char *fname,
			struct stat *st, unsigned long block0)
{
	int ret;
	int force_reads = 0;

	if (metadata_only) {
		close(fd);
		return 0;
	}

	if (total_stata > max_stata || total_open > max_stata)
		force_reads = 1;
	if (force_reads)
		run_reada(radix, reads_radix);
	if (force_reads || total_reada >= max_reada)
		run_reads(reads_radix);
	ret = add_reada_node(radix, fd, fname, st, block0);
	return ret;
}

static int read_pending_names(struct radix_tree_root *radix,
			      struct radix_tree_root *reada_radix,
			      struct radix_tree_root *reads_radix)
{
	struct node *gang[16];
	struct node *n;
	struct stat st;
	int i;
	int ret;
	int fd;
	int total_found = 0;
	while(1) {
		ret = radix_tree_gang_lookup_tag(radix,
						 (void **)gang,
						 0, ARRAY_SIZE(gang),
						 RADIX_TAG_PENDING);
		if (!ret)
			break;
		for (i = 0; i < ret; i++) {
			total_found++;
			n = gang[i];
			radix_tree_tag_clear(radix, n->ino,
					     RADIX_TAG_PENDING);
			fd = open(n->fname, O_RDONLY);
			total_open++;
			if (fd < 0)
				continue;
			ret = fstat(fd, &st);
			stat_to_dtype(n, &st);
			radix_tree_tag_clear(radix, n->ino,
					     RADIX_TAG_PENDING);
			if (n->d_type == DT_REG) {
				int ret2;
				unsigned long block0 = 0;
				if (sort_by_bmap) {
					ret2 = ioctl(fd, FIBMAP, &block0);
					if (ret2) {
						perror("fibmap ioctl");
						exit(1);
					}
				}
				if ((sort_by_bmap && block0 == 0) ||
				    (inline_files && st.st_size < 4096)) {
					struct node *p;
					read_one_fd(fd, n->fname, &st);
					p = radix_tree_delete(radix, n->ino);
					BUG_ON(p != n);
					free(n);
				} else {
					reada_one_fd(reada_radix, reads_radix,
						     fd, n->fname, &st, block0);
				}
			} else if (n->d_type == DT_DIR) {
				radix_tree_tag_set(radix, n->ino,
						   RADIX_TAG_DIR_PENDING);
				close(fd);
			}
		}
	}
	return total_found;
}

static int run_all(struct radix_tree_root *radix,
		   struct radix_tree_root *reada_radix,
		   struct radix_tree_root *reads_radix)
{
	int ret;
	int ret2;

	if (read_whole_dir_tree) {
		do {
			ret = read_next_dir(radix);
		} while (ret > 0);
	}

	do {
		ret = read_pending_names(radix, reada_radix, reads_radix);
		run_reada(reada_radix, reads_radix);
		ret2 = read_next_dir(radix);
	} while(ret > 0 || ret2 > 0);

	do {
		ret = run_reada(reada_radix, reads_radix);
		ret2 = run_reads(reads_radix);
	} while(ret > 0 || ret2 > 0);

	return 0;
}

/*
 * return seconds between start_tv and now in double precision
 */
static double time_since(struct timeval *start_tv)
{
    struct timeval stop_tv;
    double sec, usec;
    double ret;

    gettimeofday(&stop_tv, NULL);
    sec = stop_tv.tv_sec - start_tv->tv_sec;
    usec = stop_tv.tv_usec - start_tv->tv_usec;
    if (sec > 0 && usec < 0) {
        sec--;
	usec += 1000000;
    }
    ret = sec + usec / (double)1000000;
    if (ret < 0)
        ret = 0;
    return ret;
}

void print_usage() {
	fprintf(stderr, "acp version %s\n", ACP_VERSION);
	fprintf(stderr, "usage: acp [-wbvi] [-s num] [-r num] directory\n");
	fprintf(stderr, "\t-m: Metadata only, don't read the files\n");
	fprintf(stderr, "\t-w: Read entire dir tree before files\n");
	fprintf(stderr, "\t-b: Sort by bmap instead of inode number\n");
	fprintf(stderr, "\t-i: Assume small files are inline\n");
	fprintf(stderr, "\t-v: verbose output to show order things are read\n");
	fprintf(stderr, "\t-s num: max number of files to open at a time (512)\n");
	fprintf(stderr, "\t-r num: max number of MB pending in readahead() (8MB)\n");

	exit(1);
}

int main(int argc, char **argv) {
	int ret;
	int num_from;
	char *dest;
	int skip_writes = 1;
	struct radix_tree_root radix;
	struct radix_tree_root reada_radix;
	struct radix_tree_root reads_radix;
	struct stat dest_st;
	struct timeval start_time;
	double total_time;
	double mbs;
	int i;

	bufsize = 256 * 1024;
	buf = malloc(bufsize);
	if (!buf) {
		perror("malloc");
		exit(1);
	}

	INIT_RADIX_TREE(&radix, GFP_KERNEL);
	INIT_RADIX_TREE(&reada_radix, GFP_KERNEL);
	INIT_RADIX_TREE(&reads_radix, GFP_KERNEL);
	radix_tree_init();
	while(1) {
		int c;
		c = getopt(argc, argv, "s:r:Rwbivm");
		if (c < 0)
			break;
		switch(c) {
			case 'w':
				read_whole_dir_tree = 1;
				break;
			case 'm':
				metadata_only = 1;
				break;
			case 'b':
				sort_by_bmap = 1;
				break;
			case 'i':
				inline_files = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'R':
				skip_writes = 1;
				break;
			case 's':
				max_stata = atoi(optarg);
				break;
			case 'r':
				max_reada = atoi(optarg);
				max_reada = max_reada * 1024 * 1024;
				break;
			default:
				print_usage();
		}
	}
	gettimeofday(&start_time, NULL);
	num_from = argc - optind;
	if ((skip_writes && num_from < 1) || (!skip_writes && num_from < 2)) {
		print_usage();
	}
	if (!skip_writes) {
		num_from--;
		dest = argv[optind + num_from];
		ret = stat(dest, &dest_st);
		if (ret) {
			if (errno != ENOENT) {
				perror("destination stat");
				exit(1);
			}
			if (num_from > 1) {
				ret = mkdir(dest, 0777);
				if (ret) {
					perror("mkdir");
					exit(1);
				}
			}
		} else {
			if (num_from > 1 && !S_ISDIR(dest_st.st_mode)) {
				fprintf(stderr, "error: destination is not a dir\n");
				exit(1);
			}
		}
	}
	for (i = optind ; i < optind + num_from ; i++) {
		char *from;
		char *to;
		from = realpath(argv[i], NULL);
		if (!from) {
			fprintf(stderr, "path lookup failed on %s\n", argv[i]);
			exit(1);
		}
		if (skip_writes) {
			to = NULL;
		} else if (num_from > 1) {
			to = malloc(strlen(dest) + strlen(argv[i]) + 2);
			if (!to) {
				fprintf(stderr, "unable to allocate memory\n");
				exit(1);
			}
			sprintf(to, "%s/%s", dest, argv[i]);
		} else {
			to = strdup(dest);
			if (!to) {
				fprintf(stderr, "unable to allocate memory\n");
				exit(1);
			}
		}
		fprintf(stderr, "copy %s %s\n", from, to);
		ret = read_tree(&radix, from, 0);
		if (ret) {
			fprintf(stderr, "read_tree failed with "
			        "%d for %s\n", ret, argv[i]);
			exit(1);
		}
	}
	if (read_whole_dir_tree)
		fprintf(stderr, "reading entire directory tree into ram\n");
	if (sort_by_bmap)
		fprintf(stderr, "sorting by bmap instead of inode number\n");
	if (inline_files)
		fprintf(stderr, "assuming small files are inline\n");
	fprintf(stderr, "using max files open %d and max readahead %dMB\n",
		max_stata, max_reada/(1024 * 1024));
	run_all(&radix, &reada_radix, &reads_radix);
	total_time = time_since(&start_time);
	mbs = total_bytes_read / total_time;
	total_bytes_read /= 1024 * 1024;
	mbs /= 1024 * 1024;
	printf("%luMB read in %.2f seconds (%.2fMB/s)\n",
	       (unsigned long)total_bytes_read, total_time, mbs);
	return 0;
}
